/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-04
 */

#include <stdio.h>
#include "XLink.h"
#include "XName.h"

namespace nts{ // namespace nts(NiuTrans.Tensor)

int XLink::paramSize = 64;

/* constuctor */
XLink::XLink()
{
    head   = NULL;
    tails  = NULL;
    params = NULL;
    tailNum  = 0;
    paramNum = 0;
    type[0] = 0;
    typeID = 0;
}
    
/* deconstructor */
XLink::~XLink()
{
    delete[] tails;
    delete[] (char*)params;
}

/* reset it */
void XLink::Reset()
{
    delete[] tails;
    delete[] (char*)params;
    head   = NULL;
    tails  = NULL;
    params = NULL;
    tailNum  = 0;
    paramNum = 0;
    type[0]  = 0;
}

/* clear it */
void XLink::Clear()
{
    head   = NULL;
    tailNum  = 0;
    paramNum = 0;
    type[0]  = 0;
}

/* reset tails */
void XLink::ClearTail()
{
    tailNum = 0;
}

/*
clear the incoming node list of tensor node
>> node - the node to be cleared
*/
void XLink::ClearIncoming(XTensor * node)
{
    if(node == NULL)
        return;

    XLink &income = node->income;

    for(int i = 0; i < income.tailNum; i++){

        /* for a incoming node */
        XTensor * child = income.tails[i];
        XLink &childOutgo = child->outgo;

        CheckNTErrors(childOutgo.tailNum > 0, "The node must have outgoing edges!");

        /* we check for each child node and remove the link to current node */
        for(int j = 0; j < childOutgo.tailNum; j++){
            if(childOutgo.tails[j] == node){
                memcpy(childOutgo.tails + j, 
                       childOutgo.tails + j + 1, 
                      (childOutgo.tailNum - 1 - j) * sizeof(XTensor*));
                childOutgo.tailNum--;
            }
        }

        if(childOutgo.tailNum == 0)
            delete child;
    }

    income.ClearTail();
    income.tailNum = 0;
}

/* 
set edge type name 
>> id - id of the type
*/
void XLink::SetType(int id)
{
    type[0] = 0;
    strcpy(type, GetOPName(id));
    typeID = id;
    CheckNTErrors(!strcmp(type, "NULL"), "illegal edge type name!");
}

/* 
set head 
>> h - pointer to the head tensor
*/
void XLink::SetHead(XTensor * h)
{
    head = h;
}

/* 
add a tail
>> t - pointer to the tail tensor
*/
void XLink::AddTail(XTensor * t)
{
    XTensor ** ts = tails;
    tails = new XTensor*[tailNum + 1];
    memcpy(tails, ts, sizeof(XTensor*) * tailNum);
    tails[tailNum++] = t;
    delete[] ts;
}

/* 
add two tails in one time 
>> t1 - pointer to the tail tensor
>> t2 - pointer to another tail tensor
*/
void XLink::AddTwoTails(XTensor * t1, XTensor * t2)
{
    XTensor ** ts = tails;
    tails = new XTensor*[tailNum + 2];
    memcpy(tails, ts, sizeof(XTensor*) * tailNum);
    tails[tailNum++] = t1;
    tails[tailNum++] = t2;
    delete[] ts;
}

/* 
add a parameter 
>> param - parameter in default type
*/
void XLink::AddParam(DTYPE param)
{
    void * ps = params;
    params = new char[paramNum + 1];
    memcpy(params, ps, paramNum * paramSize);
    DTYPE * p = (DTYPE*)((char*)params + paramNum * paramSize);
    *p = param;
    paramNum++;
    delete[] (char*)ps;
}

/* 
add a parameter 
>> param - pointer to the parameter
>> size - size of the parameter
*/
void XLink::AddParam(void * param, int size)
{
    void * ps = params;
    params = new char[paramNum + 1];
    memcpy(params, ps, paramNum * paramSize);
    char * p = (char*)params + paramNum * paramSize;
    memcpy(p, param, size);
    paramNum++;
    delete[] (char*)ps;
}
/* 
create a hyperedge with two input tensors and a output tensor 
>> t1 - a tail tensor
>> t2 - another tail tensor
>> h - head tensor
>> id - id of the edge type
*/
void XLink::MakeLink(XTensor * t1, XTensor * t2, XTensor * h, int id)
{
    if(h != NULL)
        return;

    /* forward */
    XLink &income = h->income;
    income.Reset();
    income.SetHead(h);
    if(t1 != NULL && t2 != NULL)
        income.AddTwoTails(t1, t2);
    else if(t1 != NULL)
        income.AddTail(t1);
    else{
        ShowNTErrors("TODO!");
    }
    income.SetType(id);

    /* backward for t1 */
    if(t1 != NULL){
        XLink &outgo = t1->outgo;
        CheckNTErrors(outgo.head != t1, "Wrong head of the hyperedge!");
        outgo.AddTail(h);
    }

    /* backward for t2 */
    if(t2 != NULL){
        XLink &outgo = t2->outgo;
        CheckNTErrors(outgo.head != t2, "Wrong head of the hyperedge!");
        outgo.AddTail(h);
    }
}

/* 
create a hyper edge with a list of tensors and a output tensor 
>> list - a list of input tensors
>> h - head tensor
>> id - id of the edge type
*/
void XLink::MakeLink(XList * list, XTensor * h, int id)
{
    /* forward */
    XLink &income = h->income;
    income.Reset();
    income.SetHead(h);
    income.SetType(id);

    for(int i = 0; i < list->count; i++){
        XTensor * t = (XTensor*)list->GetItem(i);
        income.AddTail(t);
    }

    /* backward */
    for(int i = 0; i < list->count; i++){
        XTensor * t = (XTensor*)list->GetItem(i);
        XLink &outgo = t->outgo;
        CheckNTErrors(outgo.head != t, "Wrong head of the hyperedge!");
        outgo.AddTail(h);
    }
}

/* 
add parameters 
>> h - head
>> param - parameter we want introduce
*/
void XLink::AddParamToHead(XTensor * h, DTYPE param)
{
    if(h != NULL)
        return;
    h->income.AddParam(param);
}

/* 
add an integer parameter 
>> h - head
>> param - parameter we want introduce
*/
void XLink::AddParamToHeadInt(XTensor * h, int param)
{
    if(h != NULL)
        return;
    h->income.AddParam(&param, sizeof(int));
}

/* 
replace a node with another, i.e., we redirect the links to the new node 
>> oldOne - the node to be replaced
>> newOne - the new node
*/
void XLink::Replace(XTensor * oldOne, XTensor * newOne)
{
    if(oldOne == NULL || newOne == NULL)
        return;

    XLink::ClearIncoming(newOne);

    XLink &newIncome = newOne->income;
    XLink &newOutgo  = newOne->outgo;
    delete[] newIncome.tails;

    /* incoming nodes for the new node */
    newIncome.tailNum = oldOne->income.tailNum;
    newIncome.tails = new XTensor*[newIncome.tailNum];
    memcpy(newIncome.tails, oldOne->income.tails, sizeof(XTensor*) * newIncome.tailNum);

    /* update the link to each child node */
    for(int i = 0; i < newIncome.tailNum; i++){
        XTensor * child = newIncome.tails[i];
        XLink &childOutgo = child->outgo;
        for(int j = 0; j < childOutgo.tailNum; j++){
            if(childOutgo.tails[j] == oldOne){
                childOutgo.tails[j] = newOne;
            }
        }
    }

    /* outgoing nodes for the new node */
    newOutgo.tailNum = oldOne->income.tailNum;
    newOutgo.tails = new XTensor*[newOutgo.tailNum];
    memcpy(newOutgo.tails, oldOne->income.tails, sizeof(XTensor*) * newOutgo.tailNum);

    /* update the link to each parent node */
    for(int i = 0; i < newOutgo.tailNum; i++){
        XTensor * parent = newOutgo.tails[i];
        XLink &parentIncome = parent->income;
        for(int j = 0; j < parentIncome.tailNum; j++){
            if(parentIncome.tails[j] == oldOne){
                parentIncome.tails[j] = newOne;
            }
        }
    }

    XLink &oldOutgo = oldOne->outgo;
    ClearIncoming(oldOne);
    oldOne->outgo.tailNum = 0;
}
    
} // namespace nts(NiuTrans.Tensor)

